do	--Version globals

	VERSION = '10'


	Helper = {}		--useful functions
	Timer = {}		--timers
end

do	--Function groups
	Hikvision = {}
	ADDNUM = {}
	URL = {}
	RESPONSES = {}
	gTickets = {}
end

do -- setup basic commands
	CMDS = {

	PLAY = "enter",
	PAUSE = "enter",
	STOP = nil,
	SKIP_FWD = "right",
	SKIP_REV = "left",
	SCAN_REV = "down",
	SCAN_FWD = "up",
	RECORD = "record",

	NUMBER_0 = "0",
	NUMBER_1 = "1",
	NUMBER_2 = "2",
	NUMBER_3 = "3",
	NUMBER_4 = "4",
	NUMBER_5 = "5",
	NUMBER_6 = "6",
	NUMBER_7 = "7",
	NUMBER_8 = "8",
	NUMBER_9 = "9",

	STAR = "ptz",
	POUND = "preview",

	MENU = "menu",
	UP = "up",
	DOWN = "down",
	LEFT = "left",
	RIGHT = "right",
	ENTER = "enter",
	CANCEL = "esc",
	PVR = "play", -- Search

	PULSE_VOL_UP = nil,
	PULSE_VOL_DOWN = nil,
	MUTE_TOGGLE = nil,
	PULSE_CH_UP = "right",
	PULSE_CH_DOWN = "left",
	PAGE_UP = "zoom+",
	PAGE_DOWN = "zoom-1",

	-- Device specific commands
	PTZ =  "ptz",
	PREVIEW = "preview",
	A = "A",
	IRIS_IN	= "iris+",
	IRIS_OUT	= "iris=",
	FOCUS_IN = "focus+",
	FOCUS_OUT = "focus-",
	ZOOM_IN = "zoom+",
	ZOOM_OUT = "zoom-",
	F1 = "f1",
	F2 = "f2",

	-- Unsused commands for mapping

	RED = nil,
	GREEN = nil,
	YELLOW = nil,
	BLUE = nil,
	CLOSED_CAPTIONED = nil,
	SKIP_REV = nil,
	INFO = nil,
	TV_VIDEO = nil,
	GUIDE = nil,
	RECALL = nil,
	}

end

do -- Globals

	gUsername = ""
	gPassword = ""
	SettingUp = false

	AUTHCODE = ""
	PORT = 80

	CHANNELS = 16

end

function dbg (strDebugText, ...)
	if (DEBUGPRINT) then print (os.date ('%x %X : ') .. (strDebugText or ''), ...) end
end

function dbgdump (strDebugText, ...)
	if (DEBUGPRINT) then hexdump (strDebugText or '') print (...) end
end

function ExecuteCommand (strCommand, tParams)

	dbg ('Execute command: ' .. strCommand)
	if (DEBUGPRINT) then
		Helper.Print (tParams)
	end
	if (strCommand == "LUA_ACTION" and tParams and tParams.ACTION) then
		if (tParams.ACTION == 'Register') then
		end
	else -- Device specific command
		if (strCommand == 'PRESET') then
			local preset = tonumber (tParams.Preset) - 1
			cmd = CMDS['PRESET'] .. Dec2Hex (preset, 2)
		elseif (strCommand == 'CAMERA') then
			cmd = CMDS['CAMERA_'.. tParams.Camera]
		else
			cmd = CMDS[strCommand]
		end
		Hikvision.CreateMessage(cmd)
		dbg (cmd)
	end
end


function OnBindingChanged (idBinding, strClass, bIsBound)
	if (Qs and Qs [idBinding]) then
		if (bIsBound) then
			Qs [idBinding].InitialConnectTimer = Helper.AddTimer (Qs [idBinding].InitialConnectTimer, 250, 'MILLISECONDS', false)
		else
			Qs [idBinding].Disconnected ()
		end
	end
end

function OnConnectionStatusChanged (idBinding, nPort, strStatus)
	--dbg ('On Connection Status '.. idBinding ..'  ' .. nPort .. ' ' .. strStatus)

	if (idBinding == 6001) then
		if (strStatus == 'ONLINE') then
			Hikvision.InitialConnect ()
		end
	end

	if (Qs and Qs [idBinding]) then
		if (strStatus == 'ONLINE') then
			Qs [idBinding].Connected ()
		elseif (strStatus == 'OFFLINE') then
			Qs [idBinding].Disconnected ()
		end
	end
end

function OnDriverDestroyed ()
	C4:DestroyServer ()
	Helper.KillAllTimers ()
end

function OnDriverLateInit ()
	Helper.KillAllTimers ()
	Helper.DriverInfo ('')


	for k, _ in pairs (Properties) do
		OnPropertyChanged (k)
	end

	-- Reconnect if we already have a connection for driver update
	local ip = C4:GetBindingAddress(6001)
	if (ip ~= nil and string.len(ip) > 0) then OnNetworkBindingChanged(6001, true) end
end

function OnNetworkBindingChanged (idBinding, bIsBound)
	dbg (idBinding,bIsBound)
	if (idBinding == 6001) then
		if (bIsBound == true) then
			NETWORK_BOUND = true
			IP = C4:GetBindingAddress (idBinding)
			PORT = 80 -- Hard code to match XML
			--dbg('Connection bound' .. IP)
			C4:UpdateProperty ('NVR Name', 'Waiting..')
			C4:UpdateProperty ('NVR MAC', 'Waiting..')
			C4:UpdateProperty ('Serial', 'Waiting..')
			local controllerip = C4:GetControllerNetworkAddress()
			C4:SendToProxy (5001, "ADDRESS_CHANGED", {ADDRESS = IP})
			local _, _, ip_range = string.find(IP, '(%d+.%d+.%d+).%d')
			local _, _, controller_range = string.find(controllerip, '(%d+.%d+.%d+).%d')
			if (ip_range ~= controller_range) then
				Helper.DriverInfo ('WARNING: IP address not on same subnet as C4')
			else
				Helper.DriverInfo ('Attempting to connect to device')
			end
		else
			NETWORK_BOUND = false
			C4:UpdateProperty ('NVR Name', 'Setup IP Address in Connections')
			C4:UpdateProperty ('NVR MAC', ' ')
			C4:UpdateProperty ('Serial', ' ')
			Helper.DriverInfo ('Set IP on Connections tab')
			Helper.KillAllTimers ()
		end
	end
end
function OnPropertyChanged (strProperty)
	local value = Properties [strProperty]
	if (value == nil) then
		dbg ('OnPropertyChanged, nil value for Property: ', strProperty)
		return
	end

	if (strProperty == 'Automatic Updates') then
		AUTOUPDATE = (value == 'On')
		if (AUTOUPDATE and not Helper.VersionCheck21 ()) then
			Helper.DriverInfo ('Auto update requires C4 2.1 or greater')
			C4:UpdateProperty ('Automatic Updates', 'Disabled')
			OnPropertyChanged ('Automatic Updates', 'Disabled')
		end

	elseif (strProperty == 'Debug Mode') then
		if (value == 'Off') then
			DEBUGPRINT = false
			Timer.Debug = Helper.KillTimer (Timer.Debug)
			if (C4.AllowExecute) then C4:AllowExecute (false) end
		elseif (value == 'On') then
			DEBUGPRINT = true
			Timer.Debug = Helper.AddTimer (Timer.Debug, 45, 'MINUTES')
			if (C4.AllowExecute) then C4:AllowExecute (true) end
		end

	elseif (strProperty == 'Driver Version') then
		C4:UpdateProperty ('Driver Version', VERSION)

	elseif (strProperty == 'Control4 MAC Address') then
		local mac = string.upper (C4:GetUniqueMAC ())
		C4:UpdateProperty ('Control4 MAC Address', mac)

	--[[elseif (string.find (strProperty, "(%S-) .-button$"))	then
		local _, _, cmd = string.find (strProperty, "^(%S-) .-button")
		CMDS [cmd] = CMDS [value]
	]]--

	elseif (strProperty == 'Control4 MAC Address') then
		local mac = string.upper (C4:GetUniqueMAC ())
		C4:UpdateProperty ('Control4 MAC Address', mac)

	elseif (strProperty == "Username") then
		gUsername = value

	elseif (strProperty == "Password") then
		gPassword = value

	end

end

function OnServerConnectionStatusChanged (nHandle, nPort, strStatus)
	if (strStatus == 'ONLINE') then
		if (not ServerBuffer) then ServerBuffer = {} end
		ServerBuffer [nHandle] = ''
	elseif (strStatus == 'OFFLINE') then
		C4:ServerCloseClient (nHandle)
		ServerBuffer [nHandle] = nil
	end
end

function OnServerDataIn (nHandle, strData)
	if (ServerBuffer and ServerBuffer [nHandle]) then
		ServerBuffer [nHandle] = ServerBuffer [nHandle] .. strData
		C4:ServerCloseClient (nHandle)
	end
end

function OnTimerExpired (idTimer)
	if (idTimer == Timer.Debug) then
		dbg ('Turning Debug Mode Off (timer expired)')
		C4:UpdateProperty ('Debug Mode', 'Off')
		OnPropertyChanged ('Debug Mode')

	elseif (idTimer == Timer.Channel) then
		dbg ('Channel selected is '.. ChannelBuffer)
		if (ChannelBuffer ~= "") then
			if (CameraSelector == true) then
				CameraSelector = false
				if ((tonumber(ChannelBuffer) > 0) and (tonumber(ChannelBuffer) < 17)) then -- could check on max channels??
					Hikvision.CreateMessage(CMDS["CAMERA_"..ChannelBuffer])
				else
					dbg ("Camera Number > 16 entered")
				end
			elseif (WindowSelector == true) then
				WindowSelector = false
				if (ChannelBuffer == '13' or ChannelBuffer == '16') then
					Hikvision.CreateMessage(CMDS["WINDOW_"..ChannelBuffer])
				else
					dbg ("Invalid Window selection")
				end
			elseif (PresetSelector == true) then
				PresetSelector = false
				ChannelBuffer = tonumber(ChannelBuffer) - 1
				if (ChannelBuffer < 127) then
					Hikvision.CreateMessage(CMDS["PRESET"] .. Dec2Hex (ChannelBuffer, 2))
				else
					dbg ("Invalid Preset selection")
				end
			end
			ChannelBuffer = ""
		end

	elseif (idTimer == Timer.SetUp) then -- checking every 60 secs for device to be online
		Helper.DriverInfo ('No response, trying every 60 seconds, check IP address')

	elseif (idTimer == Timer.PTZ_Stop) then
		ReceivedFromProxy (5001, 'PTZ_PAN_STOP')

	else
		for _, thisQ in pairs (Qs or {}) do
			if (idTimer == thisQ.ConnectingTimer) then
				thisQ.ConnectFailed ()
			elseif (idTimer == thisQ.ConnectedTimer) then
				thisQ.ConnectStalled ()
			elseif (idTimer == thisQ.InitialConnectTimer) then
				thisQ.Connected ()
			elseif (idTimer == thisQ.MetronomeTimer) then
				thisQ.NowRTS ()
			end
		end
	end
end

function ReceivedAsync(ticketId, strData, responseCode, tHeaders)
	strData = strData or ""
	responseCode = responseCode or 0
	tHeaders = tHeaders or {}

	dbg("ReceivedAsync[" .. ticketId .. "]: Response Code: " .. responseCode .. " Length: " .. string.len(strData))
	local data_length = string.len(strData)
	if (data_length > 128) then data_length = 128 end
	dbg("ReceivedAsync[" .. ticketId .. "]: Data: " .. strData)

	if (ticketId == CheckVersionTicket) then
		Helper.CheckVersionResponse (strData, responseCode, tHeaders, strError)

	elseif (ticketId == GetMemoryUpdateTicket) then
		Helper.UpdateDriverInMemory (strData, responseCode, tHeaders, strError)

	elseif (ticketId == GetDiskUpdateTicket) then
		Helper.UpdateDriverOnDisk (strData, responseCode, tHeaders, strError)
     end


	if (responseCode ~= 200) then
		Helper.DriverInfo ('Lost communication with NVR  - could be off')
		C4:FireEvent("No response")

	elseif Properties['Driver Information'] == 'Lost communication with NVR - could be off' then
		Hikvision.InitialConnect ()
	end

	if (gTickets[ticketId] ~= nil and type(RESPONSES[gTickets[ticketId]]) == "function") then -- Need to parse response
		RESPONSES[gTickets[ticketId]](strData)
		gTickets[ticketId] = nil
	end
end
function ReceivedFromProxy (idBinding, strCommand, tParams)

	dbg (idBinding .. ' command: ' .. strCommand)
	if (DEBUGPRINT) then
		Helper.Print (tParams)
	end

	if (strCommand == "STOP") then
		C4:SendToProxy (5001, "STOP", {})
	elseif (strCommand == "PLAY") then
		C4:SendToProxy (5001, "PLAY", {})
	elseif (strCommand == "PAUSE") then
		C4:SendToProxy (5001, "PAUSE", {})
	elseif (strCommand == "PLAY") then
		C4:SendToProxy (5001, "PLAY", {})
	elseif (strCommand == "RECORD") then
		C4:SendToProxy (5001, "RECORD", {})
	end

	local cmd = CMDS [strCommand]

	local _, _, start = string.find (strCommand, 'START_(.+)')
	local _, _, stop = string.find (strCommand, 'STOP_(.+)')

	if (strCommand == 'SET_CHANNEL') then
		local channel = tonumber (tParams.CHANNEL)
		if ((channel < 1) or (channel > tonumber(CHANNELS))) then -- out of range
			dbg('Channel out of 1-' .. CHANNELS .. ' range')
			return
		else
			Hikvision.SetChannel (channel)
		end



	elseif (start) then
		if (start == 'CH_UP') then
			Hikvision.CreateMessage (CMDS ['PULSE_CH_UP'])
		elseif (start == 'CH_DOWN') then
			Hikvision.CreateMessage (CMDS ['PULSE_CH_DOWN'])
		elseif (start == 'VOL_UP') then
			Hikvision.CreateMessage (CMDS ['PULSE_VOL_UP'])
		elseif (start == 'VOL_DOWN') then
			Hikvision.CreateMessage (CMDS ['PULSE_VOL_DOWN'])
		elseif ((start == 'PAGE_UP') or (start == 'PAGE_DOWN') or (start == 'ZOOM_IN') or (start == 'ZOOM_OUT')) then
			Hikvision.CreateMessage (CMDS [start])
		elseif ((start == 'LEFT') or (start == 'RIGHT') or (start == 'UP') or (start == 'DOWN')) then
			Hikvision.CreateMessage (CMDS [start])
		end

	elseif (stop) then
		Timer [stop] = Helper.KillTimer (Timer [stop])
		local cmd = CMDS [stop]
		if (stop == 'CH_UP') then
			Hikvision.CreateMessage (CMDS ['PULSE_CH_UP'])
		elseif (stop == 'CH_DOWN') then
			Hikvision.CreateMessage (CMDS ['PULSE_CH_DOWN'])
		elseif (stop == 'VOL_UP') then
			Hikvision.CreateMessage (CMDS ['PULSE_VOL_UP'])
		elseif (stop == 'VOL_DOWN') then
			Hikvision.CreateMessage (CMDS ['PULSE_VOL_DOWN'])

		end

	elseif (cmd) then
		Hikvision.CreateMessage(cmd)

	end

end

function ReceivedFromNetwork (idBinding, nPort, strData)

	if (not Buffer) then Buffer = {} end
	if (not Buffer [idBinding]) then Buffer [idBinding] = '' end
	dbg ("Received data on idBinding" .. idBinding)
	--dbgdump(strData)
	Buffer [idBinding] = Buffer [idBinding] .. strData


	if (Qs and Qs [idBinding]) then
		Qs [idBinding].Receiving ()
		if (Qs [idBinding].PERMANENT) then
			local rts
			--parsing functions here to check for complete message worth parsing
			if (rts) then
				Qs [idBinding].NowRTS ()
			end
		elseif (Qs [idBinding].TRANSIENT and Qs [idBinding].UDP and Qs [idBinding].ProcessMessage) then	--needed as UDP doesn't go offline even in transient Q
			Qs [idBinding].ProcessMessage ()
		end
	end
end

function ReceivedFromSerial (idBinding, strData)

	if (not Buffer) then Buffer = {} end
	if (not Buffer [idBinding]) then Buffer [idBinding] = '' end

	Buffer [idBinding] = Buffer [idBinding] .. strData

	if (Qs and Qs [idBinding]) then
		Qs [idBinding].Receiving ()
		local rts
		dbg(Buffer[idBinding])
		--parsing functions here to check for complete message worth parsing
		if (rts) then
			Qs [idBinding].NowRTS ()
		end
	end
end

function Helper.AddTimer (timer, count, units, recur)
	local newTimer
	if (recur == nil) then recur = false end
	if (timer and timer ~= 0) then Helper.KillTimer (timer) end

	newTimer = C4:AddTimer (count, units, recur)
	return newTimer
end

function Helper.Deflate (input)

	local function warn(s)
		dbg (s, '\n')
	end


	local function make_outstate(outbs)
		local outstate = {}
		outstate.outbs = outbs
		outstate.window = {}
		outstate.window_pos = 1
		return outstate
	end


	local function output(outstate, byte)
		local window_pos = outstate.window_pos
		outstate.outbs(byte)
		outstate.window [window_pos] = byte
		outstate.window_pos = window_pos % 32768 + 1	-- 32K
	end


	local function noeof(val)
		return (val)
	end


	local function hasbit(bits, bit)
		return bits % (bit + bit) >= bit
	end


	local function memoize(f)
		local mt = {}
		local t = setmetatable({}, mt)
		function mt:__index(k)
		local v = f(k)
		t [k] = v
		return v
		end
		return t
	end


	-- small optimization (lookup table for powers of 2)
	local pow2 = memoize(function(n) return 2^n end)

	--local tbits = memoize(
	--	function(bits)
	--		return memoize( function(bit) return getbit(bits, bit) end )
	--	end )


	-- weak metatable marking objects as bitstream type
	local is_bitstream = setmetatable({}, {__mode='k'})


	-- DEBUG
	-- prints LSB first
	local function bits_tostring(bits, nbits)
		local s = ''
		local tmp = bits
		local function f()
		local b = tmp % 2 == 1 and 1 or 0
		s = s .. b
		tmp = (tmp - b) / 2
		end
		if nbits then
		for i=1,nbits do f() end
		else
		while tmp ~= 0 do f() end
		end

		return s
	end


	local function bytestream_from_file(fh)
		local o = {}
		function o:read()
		local sb = fh:read(1)
		if sb then return sb:byte() end
		end
		return o
	end


	local function bytestream_from_string(s)
		local i = 1
		local o = {}
		function o:read()
		local by
		if i <= #s then
			by = s:byte(i)
			i = i + 1
		end
		return by
		end
		return o
	end

	local left
	local function bitstream_from_bytestream(bys)
		local buf_byte = 0
		local buf_nbit = 0
		local o = {}

		function o:nbits_left_in_byte()
		return buf_nbit
		end

		function o:read(nbits)
		nbits = nbits or 1
		while buf_nbit < nbits do
			local byte = bys:read()
			if not byte then return end	-- note: more calls also return nil
			buf_byte = buf_byte + pow2 [buf_nbit] * byte
			buf_nbit = buf_nbit + 8
		end
		local m = pow2 [nbits]
		local bits = buf_byte % m
		buf_byte = (buf_byte - bits) / m
		buf_nbit = buf_nbit - nbits
		return bits
		end

		is_bitstream [o] = true

		return o
	end


	local function get_bitstream(o)
		local bs
		if type(o) == 'string' then
		bs = bitstream_from_bytestream(bytestream_from_string(o))
		elseif is_bitstream [o] then
		return o
		else
		return 'unrecognized type'
		end
		return bs
	end


	local function get_obytestream(o)
		local bs
		if type(o) == 'function' then
		bs = o
		else
		dbg ('unrecognized type: ' .. tostring(o))
		end
		return bs
	end


	local function HuffmanTable(init, is_full)
		local t = {}
		if is_full then
		for val,nbits in pairs(init) do
			if nbits ~= 0 then
			t [#t+1] = {val=val, nbits=nbits}
			end
		end
		else
		for i=1,#init-2,2 do
			local firstval, nbits, nextval = init [i], init [i+1], init [i+2]
			if nbits ~= 0 then
			for val=firstval,nextval-1 do
				t [#t+1] = {val=val, nbits=nbits}
			end
			end
		end
		end
		table.sort(t, function(a,b)
		return a.nbits == b.nbits and a.val < b.val or a.nbits < b.nbits
		end)

		-- assign codes
		local code = 1	-- leading 1 marker
		local nbits = 0
		for i,s in ipairs(t) do
		if s.nbits ~= nbits then
			code = code * pow2 [s.nbits - nbits]
			nbits = s.nbits
		end
		s.code = code
		code = code + 1
		end

		local minbits = math.huge
		local look = {}
		for i,s in ipairs(t) do
		minbits = math.min(minbits, s.nbits)
		look [s.code] = s.val
		end

		-- function t:lookup(bits) return look [bits] end

		local function msb(bits, nbits)
		local res = 0
		for i=1,nbits do
			local b = bits % 2
			bits = (bits - b) / 2
			res = res * 2 + b
		end
		return res
		end
		local tfirstcode = memoize(
		function(bits) return pow2 [minbits] + msb(bits, minbits) end)

		function t:read(bs)
		local code = 1 -- leading 1 marker
		local nbits = 0
		while 1 do
			if nbits == 0 then	-- small optimization (optional)
			code = tfirstcode [noeof(bs:read(minbits))]
			nbits = nbits + minbits
			else
			local b = noeof(bs:read())
			nbits = nbits + 1
			code = code * 2 + b	-- MSB first
			end
			local val = look [code]
			if val then
			return val
			end
		end
		end

		return t
	end


	local function parse_gzip_header(bs)
		-- local FLG_FTEXT = 2^0
		local FLG_FHCRC = 2^1
		local FLG_FEXTRA = 2^2
		local FLG_FNAME = 2^3
		local FLG_FCOMMENT = 2^4

		local id1 = bs:read(8)
		local id2 = bs:read(8)
		if id1 ~= 31 or id2 ~= 139 then
		return 'not in gzip format'
		end
		local cm = bs:read(8)	-- compression method
		local flg = bs:read(8) -- FLaGs
		local mtime = bs:read(32) -- Modification TIME
		local xfl = bs:read(8) -- eXtra FLags
		local os = bs:read(8) -- Operating System

		if not os then return 'invalid header' end

		if hasbit(flg, FLG_FEXTRA) then
		local xlen = bs:read(16)
		local extra = 0
		for i=1,xlen do
			extra = bs:read(8)
		end
		if not extra then return 'invalid header' end
		end

		local function parse_zstring(bs)
		repeat
			local by = bs:read(8)
			if not by then return 'invalid header' end
		until by == 0
		end

		if hasbit(flg, FLG_FNAME) then
		parse_zstring(bs)
		end

		if hasbit(flg, FLG_FCOMMENT) then
		parse_zstring(bs)
		end

		if hasbit(flg, FLG_FHCRC) then
		local crc16 = bs:read(16)
		if not crc16 then return 'invalid header' end
		end
	end


	local function parse_huffmantables(bs)
		local hlit = bs:read(5)	-- # of literal/length codes - 257
		local hdist = bs:read(5) -- # of distance codes - 1
		local hclen = noeof(bs:read(4)) -- # of code length codes - 4

		local ncodelen_codes = hclen + 4
		local codelen_init = {}
		local codelen_vals = {
			16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15}
		for i=1,ncodelen_codes do
			local nbits = bs:read(3)
			local val = codelen_vals [i]
			codelen_init [val] = nbits
		end
		local codelentable = HuffmanTable(codelen_init, true)

		local function decode(ncodes)
			local init = {}
			local nbits
			local val = 0
			while val < ncodes do
			local codelen = codelentable:read(bs)
			--FIX:check nil?
			local nrepeat
			if codelen <= 15 then
				nrepeat = 1
				nbits = codelen
			elseif codelen == 16 then
				nrepeat = 3 + noeof(bs:read(2))
				-- nbits unchanged
			elseif codelen == 17 then
				nrepeat = 3 + noeof(bs:read(3))
				nbits = 0
			elseif codelen == 18 then
				nrepeat = 11 + noeof(bs:read(7))
				nbits = 0
			else
				dbg 'ASSERT'
			end
			for i=1,nrepeat do
				init [val] = nbits
				val = val + 1
			end
			end
			local huffmantable = HuffmanTable(init, true)
			return huffmantable
		end

		local nlit_codes = hlit + 257
		local ndist_codes = hdist + 1

		local littable = decode(nlit_codes)
		local disttable = decode(ndist_codes)

		return littable, disttable
	end


	local tdecode_len_base
	local tdecode_len_nextrabits
	local tdecode_dist_base
	local tdecode_dist_nextrabits
	local function parse_compressed_item(bs, outstate, littable, disttable)
		local val = littable:read(bs)
		if val < 256 then -- literal
		output(outstate, val)
		elseif val == 256 then -- end of block
		return true
		else
		if not tdecode_len_base then
			local t = {[257] = 3}
			local skip = 1
			for i=258,285,4 do
			for j=i,i+3 do t [j] = t [j-1] + skip end
			if i ~= 258 then skip = skip * 2 end
			end
			t [285] = 258
			tdecode_len_base = t
		end
		if not tdecode_len_nextrabits then
			local t = {}
			for i=257,285 do
			local j = math.max(i - 261, 0)
			t [i] = (j - (j % 4)) / 4
			end
			t [285] = 0
			tdecode_len_nextrabits = t
		end
		local len_base = tdecode_len_base [val]
		local nextrabits = tdecode_len_nextrabits [val]
		local extrabits = bs:read(nextrabits)
		local len = len_base + extrabits

		if not tdecode_dist_base then
			local t = {[0] = 1}
			local skip = 1
			for i=1,29,2 do
			for j=i,i+1 do t [j] = t [j-1] + skip end
			if i ~= 1 then skip = skip * 2 end
			end
			tdecode_dist_base = t
		end
		if not tdecode_dist_nextrabits then
			local t = {}
			for i=0,29 do
			local j = math.max(i - 2, 0)
			t [i] = (j - (j % 2)) / 2
			end
			tdecode_dist_nextrabits = t
		end
		local dist_val = disttable:read(bs)
		local dist_base = tdecode_dist_base [dist_val]
		local dist_nextrabits = tdecode_dist_nextrabits [dist_val]
		local dist_extrabits = bs:read(dist_nextrabits)
		local dist = dist_base + dist_extrabits

		for i=1,len do
			local pos = (outstate.window_pos - 1 - dist) % 32768 + 1	-- 32K
			if (outstate.window [pos]) then
			output(outstate, outstate.window [pos])
			else
				return (nil)
			end
		end
		end
		return false
	end


	local function parse_block(bs, outstate)
		local bfinal = bs:read(1)
		local btype = bs:read(2)

		local BTYPE_NO_COMPRESSION = 0
		local BTYPE_FIXED_HUFFMAN = 1
		local BTYPE_DYNAMIC_HUFFMAN = 2
		local BTYPE_RESERVED = 3

		if btype == BTYPE_NO_COMPRESSION then
		bs:read(bs:nbits_left_in_byte())
		local len = bs:read(16)
		local nlen = noeof(bs:read(16))

		for i=1,len do
			local by = noeof(bs:read(8))
			output(outstate, by)
		end
		elseif btype == BTYPE_FIXED_HUFFMAN or btype == BTYPE_DYNAMIC_HUFFMAN then
		local littable, disttable
		if btype == BTYPE_DYNAMIC_HUFFMAN then
			littable, disttable = parse_huffmantables(bs)
		else
			littable	= HuffmanTable {0,8, 144,9, 256,7, 280,8, 288,nil}
			disttable = HuffmanTable {0,5, 32,nil}
		end

		repeat
			local is_done = parse_compressed_item(
			bs, outstate, littable, disttable)
		until is_done
		else
		return 'unrecognized compression type'
		end

		return bfinal ~= 0
	end

	local ostring = ''
	local function getostring (sbyte)
		ostring = ostring .. string.char (sbyte)
	end

	local function deflate(t)
		local bs = get_bitstream(t)
		local outbs = getostring
		local outstate = make_outstate(outbs)

		repeat
		local is_final = parse_block(bs, outstate)
		until is_final
	end

	local function gunzip(t)
		local bs = get_bitstream(t)
		local outbs = getostring

		if (parse_gzip_header(bs) == 'not in gzip format') then
			ostring = input
			return
		end

		deflate (bs)

		bs:read(bs:nbits_left_in_byte())

		local crc32 = bs:read(32)
		local isize = bs:read(32) -- ignored
	end

	gunzip (input)

	return ostring
end

function Helper.DriverInfo (info)
	C4:UpdateProperty ('Driver Information', info)
	print (os.date ('%x %X : ') .. info)
end

function Helper.FileRead (filename)
	local content = ''
	if (C4:FileExists (filename)) then
		local file = C4:FileOpen (filename)
		local length = C4:FileGetSize (file)
		C4:FileSetPos (file, 0)
		content = C4:FileRead (file, length)
		C4:FileClose (file)
	end
	return (content)
end

function Helper.FileWrite (filename, content, overwrite)
	content = tostring (content) or ''
	local pos = 0
	if (overwrite and C4:FileExists (filename)) then
		C4:FileDelete (filename)
	end
	local file = C4:FileOpen (filename)
	if (not overwrite) then
		pos = C4:FileGetSize (file)
	end
	C4:FileSetPos (file, pos)
	local result = C4:FileWrite (file, content:len (), content)
	C4:FileClose (file)
end

function Helper.GetControllerType ()
	C4:FileSetDir ('/etc')
	local motd = Helper.FileRead ('motd')
	C4:FileSetDir ('')
	if (string.find (motd, '1000')) then
		CONTROLLERTYPE = 1000
	elseif (string.find (motd, '800')) then
		CONTROLLERTYPE = 800
	elseif (string.find (motd, '500')) then
		CONTROLLERTYPE = 500
	elseif (string.find (motd, '300')) then
		CONTROLLERTYPE = 300
	elseif (string.find (motd, '250')) then
		CONTROLLERTYPE = 250
	elseif (string.find (motd, '200')) then
		CONTROLLERTYPE = 200
	else
		CONTROLLERTYPE = 1
	end
end

function Helper.GetLocalIPAddress ()
	if (LuaC4Object.GetControllerNetworkAddress and C4:GetControllerNetworkAddress () ~= '') then
		return C4:GetControllerNetworkAddress ()
	else
		local function getIP (interface)
			local tcp = socket.tcp ()
			tcp:connect ('127.0.0.1', 5800)
			tcp:settimeout (0.5)
			tcp:send ('net iface ' .. interface .. ' real\r\n')
			local text = ''

			repeat
				local value = tcp:receive ()
				text = text .. (value or '') .. '\r\n'

			until (value == nil)

			local address = string.match (text, 'address=(.-)\r\n')
			local netmask = string.match (text, 'netmask=(.-)\r\n')
			local gateway = string.match (text, 'gateway=(.-)\r\n')
			local dhcp = (string.find (text, 'type=dhcp') ~= nil)
			tcp:close ()
			return address, netmask, gateway, dhcp
		end

		local address, netmask, gateway, dhcp = getIP ('eth0')
		if (not (address)) then
			address, netmask, gateway, dhcp = getIP ('wlan0')
		end

		return address, netmask, gateway, dhcp
	end
end

function Helper.KillAllTimers ()

	for k,v in pairs (Timer or {}) do
		if (type (v) == 'number') then
			Timer [k] = Helper.KillTimer (Timer [k])
		end
	end

	for _, thisQ in pairs (Qs or {}) do
		if (thisQ.ConnectingTimer and thisQ.ConnectingTimer ~= 0) then thisQ.ConnectingTimer = Helper.KillTimer (thisQ.ConnectingTimer) end
		if (thisQ.ConnectedTimer and thisQ.ConnectedTimer ~= 0) then thisQ.ConnectedTimer = Helper.KillTimer (thisQ.ConnectedTimer) end
	end
end

function Helper.KillTimer (timer)
	if (timer and type (timer) == 'number') then
		return (C4:KillTimer (timer))
	else
		return (0)
	end
end

function Helper.MakeSerialQ (idBinding, tParams)
	local Q = {}
	Q.PERMANENT = true

	if (not idBinding) then return end

	if (not (Qs and type (Qs) == 'table')) then Qs = {} end
	if (Qs [idBinding] and Qs [idBinding].KillQ) then Qs [idBinding].KillQ () end
	Qs [idBinding] = Q

	if (not (Buffer and type (Buffer) == 'table')) then Buffer = {} end
	Buffer [idBinding] = ''

	local connectedTimeout, terminator, maxFailCount, name, limit, metronome = 5000, '', 3, idBinding, 50, 0

	if (tParams and type (tParams) == 'table') then
		if (tParams.CONNECTED) then connectedTimeout = tonumber (tParams.CONNECTED) or connectedTimeout end
		if (tParams.TERMINATOR) then terminator = tParams.TERMINATOR end
		if (tParams.FAILCOUNT) then maxFailCount = tonumber (tParams.FAILCOUNT) or 3 end
		if (tParams.NAME) then name = tostring (tParams.NAME) or idBinding end
		if (tParams.LIMIT) then limit = tonumber (tParams.LIMIT) or 50 end
		if (tParams.METRONOME) then metronome = tonumber (tParams.METRONOME) or 0 end
	end

	Q.BuildConnection = function ()
		Buffer [idBinding] = ''
		Q.ClearQueue ()

		Q.NetConnected = (C4:GetBoundProviderDevice (C4:GetDeviceID (), idBinding) ~= 0)
		Q.RTS = Q.NetConnected
		if (Q.RTS and Q.Initialize) then
			Q.Initialize ()
		end

		Q.ConnectedTimer = 0
		Q.FailCount = 0
		Q.BINDING = idBinding
	end

	Q.AddMessage = function (message)
		if (table.getn (Q) > limit) then
			dbg ('EVQ - Queue limit reached, resetting connection ' .. name)
			Q.ConnectStalled ()
		else
			table.insert (Q, message .. terminator)
			Q.SendMessage ()
		end
	end

	Q.SendMessage = function ()
		if (Q.RTS) then
			if (table.getn (Q) > 0) then
				local message = table.remove (Q, 1)
				if (message) then
					Q.RTS = false
					C4:SendToSerial (idBinding, message)
					Q.ConnectedTimer = Helper.AddTimer (Q.ConnectedTimer, connectedTimeout, 'MILLISECONDS', false)
				end
				if (metronome ~= 0) then -- tick along
					Q.MetronomeTimer = Helper.AddTimer (Q.MetronomeTimer, metronome,  'MILLISECONDS', false)
				end
			end
		end
	end

	Q.Connected = function ()
		Q.NetConnected = true
		Q.FailCount = 0
		Q.NowRTS ()
		if (Q.Initialize) then
			Q.Initialize ()
		end
	end

	Q.Receiving = function ()
		if (Q.ConnectedTimer ~= 0) then
			Q.ConnectedTimer = Helper.AddTimer (Q.ConnectedTimer, connectedTimeout, 'MILLISECONDS', false)
		end
	end

	Q.Disconnected = function ()
		Q.NetConnected = false
		Q.ClearQueue ()
		Q.RTS = false
		Q.ConnectedTimer = Helper.KillTimer (Q.ConnectedTimer)
	end

	Q.NowRTS = function ()
		Q.ConnectedTimer = Helper.KillTimer (Q.ConnectedTimer)
		Q.RTS = true
		Q.SendMessage ()
	end

	Q.ClearQueue = function ()
		while (table.getn (Q) > 0) do
			table.remove (Q)
		end
	end

	Q.ConnectFailed = function ()
		Q.ConnectStalled ()
	end

	Q.ConnectStalled = function ()
		Q.FailCount = Q.FailCount + 1
		dbg ('EVQ - Timeout while connected - Count: ' .. Q.FailCount .. ' on connection ' .. name)
		if (Q.FailCount > maxFailCount) then
			Q.ClearQueue ()
			Q.ConnectedTimer = Helper.KillTimer (Q.ConnectedTimer)
			Q.BuildConnection ()
			if (Q.ConnectStall) then
				Q.ConnectStall ()
			end
		else
			Q.NowRTS ()
		end
	end

	Q.KillQ = function ()
		Qs[idBinding] = nil
		return nil
	end

	Q.BuildConnection ()

	return Q
end

function Helper.MakePermanentNetQ (idBinding, address, port, tParams)
	local Q = {}
	Q.PERMANENT = true
	if (not idBinding) then
		repeat
			idBinding = math.random (6100, 6999)
			if (Qs[idBinding]) then idBinding = nil end
		until (idBinding ~= nil)
	end

	if (not address and port) then return end

	if (not (Qs and type (Qs) == 'table')) then Qs = {} end
	if (Qs [idBinding] and Qs [idBinding].KillQ) then Qs [idBinding].KillQ () end
	Qs [idBinding] = Q

	if (not (Buffer and type (Buffer) == 'table')) then Buffer = {} end
	Buffer [idBinding] = ''

	local SSL, UDP, connectedTimeout, connectingTimeout, terminator, maxFailCount, name, limit, metronome, xml_connection = false, false, 5000, 5000, '', 3, idBinding, 50, 0, false

	if (tParams and type (tParams) == 'table') then
		if (tParams.SSL) then SSL = true Q.SSL = true end
		if (tParams.UDP) then UDP = true Q.UDP = true end
		if (tParams.CONNECTED) then connectedTimeout = tonumber (tParams.CONNECTED) or connectedTimeout end
		if (tParams.CONNECTING) then connectingTimeout = tonumber (tParams.CONNECTING) or connectingTimeout end
		if (tParams.TERMINATOR) then terminator = tParams.TERMINATOR end
		if (tParams.FAILCOUNT) then maxFailCount = tonumber (tParams.FAILCOUNT) or 3 end
		if (tParams.NAME) then name = tostring (tParams.NAME) or idBinding end
		if (tParams.LIMIT) then limit = tonumber (tParams.LIMIT) or 50 end
		if (tParams.METRONOME) then metronome = tonumber (tParams.METRONOME) or 0 end
		if (tParams.XML_CONNECTION) then xml_connection = true end
	end

	Q.BuildConnection = function ()
		Buffer [idBinding] = ''
		Q.ClearQueue ()

		if (xml_connection == false) then
			if (SSL) then
				C4:CreateNetworkConnection (idBinding, address, 'SSL')
			else
				C4:CreateNetworkConnection (idBinding, address)
			end
		end

		Q.ConnectingTimer = 0
		Q.ConnectedTimer = 0
		Q.FailCount = 0
		Q.NetConnected = false
		Q.NetConnecting = false
		Q.BINDING = idBinding
		Q.ADDRESS = address
		Q.PORT = port

		if (Q.PERMANENT) then
			Q.RTS = false
			if (UDP) then
				C4:NetConnect (idBinding, port, 'UDP')
			else
				C4:NetConnect (idBinding, port)
			end

			Q.ConnectingTimer = Helper.AddTimer (Q.ConnectingTimer, connectingTimeout, 'MILLISECONDS', false)
		end
	end

	Q.AddMessage = function (message)
		if (table.getn (Q) > limit) then
			dbg ('NetQ - Queue limit reached, resetting connection ' .. name)
			Q.ConnectFailed ()
		else
			table.insert (Q, message .. terminator)
			Q.SendMessage ()
		end
	end

	Q.Receiving = function ()
		if (Q.ConnectedTimer ~= 0) then
			Q.ConnectedTimer = Helper.AddTimer (Q.ConnectedTimer, connectedTimeout, 'MILLISECONDS', false)
		end
	end

	Q.ClearQueue = function ()
		Buffer [idBinding] = ''
		while (table.getn (Q) > 0) do
			table.remove (Q)
		end
	end

	Q.KillQ = function ()
		C4:NetDisconnect (idBinding, port)
		Q.ConnectedTimer = Helper.KillTimer (Q.ConnectedTimer)
		Q.ConnectingTimer = Helper.KillTimer (Q.ConnectingTimer)
		Qs[idBinding] = nil
		Buffer [idBinding] = nil
		return nil
	end

	Q.SendMessage = function ()
		if (Q.RTS) then
			if (table.getn (Q) > 0) then
				local message = table.remove (Q, 1)
				if (message) then
					Q.RTS = false
					C4:SendToNetwork (idBinding, port, message)
					Q.ConnectedTimer = Helper.AddTimer (Q.ConnectedTimer, connectedTimeout, 'MILLISECONDS', false)
				end
				if (metronome ~= 0) then -- tick along
					Q.MetronomeTimer = Helper.AddTimer (Q.MetronomeTimer, metronome,  'MILLISECONDS', false)
				end
			end
		end
	end

	Q.Connected = function ()
		Q.ConnectingTimer = Helper.KillTimer (Q.ConnectingTimer)
		Q.FailCount = 0
		Q.NowRTS ()
	end

	Q.Disconnected = function ()
		Q.ConnectedTimer = Helper.KillTimer (Q.ConnectedTimer)
		Q.ConnectingTimer = Helper.AddTimer (Q.ConnectingTimer, connectingTimeout, 'MILLISECONDS', false)
	end

	Q.NowRTS = function ()
		Q.ConnectedTimer = Helper.KillTimer (Q.ConnectedTimer)
		Q.RTS = true
		Q.SendMessage ()
	end

	Q.ConnectFailed = function ()
		C4:NetDisconnect (idBinding, port)
		dbg ('EVQ - Failed to connect on connection ' .. name .. ': check IP address')
		Q.ClearQueue ()
		Q.ConnectingTimer = Helper.KillTimer (Q.ConnectingTimer)
		Q.ConnectedTimer = Helper.KillTimer (Q.ConnectedTimer)
		Q.BuildConnection ()
		if (Q.ConnectFail) then
			Q.ConnectFail ()
		end
	end

	Q.ConnectStalled = function ()
		Q.FailCount = Q.FailCount + 1
		Q.NetConnecting = false
		Q.NetConnected = false
		dbg ('NetQ - Timeout while connected - Count: ' .. Q.FailCount .. ' on connection ' .. name)
		if (Q.FailCount > maxFailCount) then
			Q.ClearQueue ()
			Q.ConnectingTimer = Helper.KillTimer (Q.ConnectingTimer)
			Q.ConnectedTimer = Helper.KillTimer (Q.ConnectedTimer)
			Q.BuildConnection ()
			if (Q.ConnectStall) then
				Q.ConnectStall ()
			end
		else
			Q.NowRTS ()
		end
	end

	Q.BuildConnection ()

	return Q
end


function Helper.MakeTransientNetQ (idBinding, address, port, tParams)
	local Q = {}
	Q.TRANSIENT = true
	if (not idBinding) then
		repeat
			idBinding = math.random (6100, 6999)
			if (Qs[idBinding]) then idBinding = nil end
		until (idBinding ~= nil)
	end

	if (not address and port) then return end

	if (not (Qs and type (Qs) == 'table')) then Qs = {} end
	if (Qs [idBinding] and Qs [idBinding].KillQ) then Qs [idBinding].KillQ () end
	Qs [idBinding] = Q

	if (not (Buffer and type (Buffer) == 'table')) then Buffer = {} end
	Buffer [idBinding] = ''

	local SSL, UDP, connectedTimeout, connectingTimeout, terminator, maxFailCount, name, limit, metronome, xml_connection = false, false, 5000, 5000, '', 3, idBinding, 50, 0, false

	if (tParams and type (tParams) == 'table') then
		if (tParams.SSL) then SSL = true Q.SSL = true end
		if (tParams.UDP) then UDP = true Q.UDP = true end
		if (tParams.CONNECTED) then connectedTimeout = tonumber (tParams.CONNECTED) or connectedTimeout end
		if (tParams.CONNECTING) then connectingTimeout = tonumber (tParams.CONNECTING) or connectingTimeout end
		if (tParams.TERMINATOR) then terminator = tParams.TERMINATOR end
		if (tParams.FAILCOUNT) then maxFailCount = tonumber (tParams.FAILCOUNT) or 3 end
		if (tParams.NAME) then name = tostring (tParams.NAME) or idBinding end
		if (tParams.LIMIT) then limit = tonumber (tParams.LIMIT) or 50 end
		if (tParams.METRONOME) then metronome = tonumber (tParams.METRONOME) or 0 end
		if (tParams.XML_CONNECTION) then xml_connection = true end
	end

	Q.BuildConnection = function ()
		Buffer [idBinding] = ''
		Q.ClearQueue ()

		if (xml_connection == false) then
			if (SSL) then
				C4:CreateNetworkConnection (idBinding, address, 'SSL')
			else
				C4:CreateNetworkConnection (idBinding, address)
			end
		end

		Q.ConnectingTimer = 0
		Q.ConnectedTimer = 0
		Q.FailCount = 0
		Q.NetConnected = false
		Q.NetConnecting = false
		Q.BINDING = idBinding
		Q.ADDRESS = address
		Q.PORT = port

		if (Q.PERMANENT) then
			Q.RTS = false
			if (UDP) then
				C4:NetConnect (idBinding, port, 'UDP')
			else
				C4:NetConnect (idBinding, port)
			end

			Q.ConnectingTimer = Helper.AddTimer (Q.ConnectingTimer, connectingTimeout, 'MILLISECONDS', false)
		end
	end

	Q.AddMessage = function (message)
		if (table.getn (Q) > limit) then
			dbg ('EVQ - Queue limit reached, resetting connection ' .. name)
			Q.ConnectFailed ()
		else
			table.insert (Q, message .. terminator)
			Q.SendMessage ()
		end
	end

	Q.Receiving = function ()
		if (Q.ConnectedTimer ~= 0) then
			Q.ConnectedTimer = Helper.AddTimer (Q.ConnectedTimer, connectedTimeout, 'MILLISECONDS', false)
		end
	end

	Q.ClearQueue = function ()
		Buffer [idBinding] = ''
		while (table.getn (Q) > 0) do
			table.remove (Q)
		end
	end

	Q.KillQ = function ()
		C4:NetDisconnect (idBinding, port)
		Q.ConnectedTimer = Helper.KillTimer (Q.ConnectedTimer)
		Q.ConnectingTimer = Helper.KillTimer (Q.ConnectingTimer)
		Qs[idBinding] = nil
		Buffer [idBinding] = nil
		return nil
	end

	Q.SendMessage = function ()
		if (not (Q.NetConnecting or Q.NetConnected)) then
			if (table.getn(Q) > 0) then
				Q.NetConnecting = true
				if (UDP) then
					C4:NetConnect (idBinding, port, 'UDP')
				else
					C4:NetConnect (idBinding, port)
				end
				Q.ConnectingTimer = Helper.AddTimer (Q.ConnectingTimer, connectingTimeout, 'MILLISECONDS', false)
			end
		end
	end

	Q.Connected = function ()
		Q.ConnectingTimer = Helper.KillTimer (Q.ConnectingTimer)
		Q.ConnectedTimer = Helper.AddTimer (Q.ConnectedTimer, connectedTimeout, 'MILLISECONDS', false)
		Q.FailCount = 0
		Q.NetConnected = true
		Q.NetConnecting = false
		local message = table.remove (Q, 1)
		if (message) then
			C4:SendToNetwork (idBinding, port, message)
		else
			C4:NetDisconnect (idBinding, port)
		end
	end

	Q.Disconnected = function ()
		Q.ConnectedTimer = Helper.KillTimer (Q.ConnectedTimer)
		if (Q.NetConnecting) then
			Q.ConnectingTimer = Helper.KillTimer (Q.ConnectingTimer)
			dbg ('EVQ - Offline unexpectedly on connection ' .. name)
			Q.ConnectFailed ()
		else
			if (Q.ProcessMessage) then
				Q.ProcessMessage ()
			end
			Q.NetConnected = false
			Q.SendMessage ()
		end
	end

	Q.ConnectFailed = function ()
		Q.FailCount = Q.FailCount + 1
		Q.NetConnecting = false
		Q.NetConnected = false
		C4:NetDisconnect (idBinding, port)
		dbg ('EVQ - Failed to connect - Count: ' .. Q.FailCount .. ' on connection ' .. name .. ': check IP address')
		if (Q.FailCount > maxFailCount) then
			Q.ClearQueue ()
			Q.ConnectingTimer = Helper.KillTimer (Q.ConnectingTimer)
			Q.ConnectedTimer = Helper.KillTimer (Q.ConnectedTimer)
			Q.BuildConnection ()
			if (Q.ConnectFail) then
				Q.ConnectFail ()
			end
		else
			Q.SendMessage ()
		end
	end

	Q.ConnectStalled = function ()
		Q.FailCount = Q.FailCount + 1
		Q.NetConnecting = false
		Q.NetConnected = false
		C4:NetDisconnect (idBinding, port)
		dbg ('EVQ - Timeout while connected - Count: ' .. Q.FailCount .. ' on connection ' .. name)
		if (Q.FailCount > maxFailCount) then
			Q.ClearQueue ()
			Q.ConnectingTimer = Helper.KillTimer (Q.ConnectingTimer)
			Q.ConnectedTimer = Helper.KillTimer (Q.ConnectedTimer)
			Q.BuildConnection ()
			if (Q.ConnectStall) then
				Q.ConnectStall ()
			end
		else
			Q.SendMessage ()
		end
	end

	Q.BuildConnection ()

	return Q
end

function Helper.Print (data)
	if (type (data) == 'table') then
		for k, v in pairs (data) do print (k, v) end
	elseif (type (data) ~= 'nil') then
		print (type (data), data)
	else
		print ('nil value')
	end
end

function Helper.ProcessHTTP (idBinding, isServer)

	local Buffer = Buffer
	if (isServer) then Buffer = ServerBuffer end

	local i, _, HTTP_Header, HTTP_Content = string.find (Buffer [idBinding], '(.-\r\n)\r\n(.*)')

	if (i == nil) then	-- no <cr><lf><cr><lf> received yet
		dbg ("Message incomplete")
		return
	end

	local rcvHeaders = {}

	string.gsub (HTTP_Header, '([^%s:]+):?%s?([^\r\n]*)\r\n', function (a, b) rcvHeaders [string.upper (a)] = b return '' end)

	if (rcvHeaders ['CONTENT-LENGTH'] and HTTP_Content:len() < tonumber (rcvHeaders ['CONTENT-LENGTH'])) then	-- Packet not complete yet, still waiting for content to arrive
		return
	end

	if (rcvHeaders ['TRANSFER-ENCODING'] and string.find (rcvHeaders ['TRANSFER-ENCODING'], 'chunked')) then	--what about a stream? Do we need to deal with that?
		local result = ''
		local lastChunk = false

		while (not lastChunk) do
			local _, chunkStart, chunkLen = string.find (HTTP_Content, '^%W-(%x+).-\r\n')	--allow for chunk extensions between HEX and crlf
			if (not chunkLen) then return end --waiting for more HTTP chunks  	--waiting for more to come in?
			chunkLen = tonumber (chunkLen, 16)

			if (chunkLen == 0 or chunkLen == nil) then
				lastChunk = true
			else
				result = result .. string.sub (HTTP_Content, chunkStart + 1, chunkStart + chunkLen)

				HTTP_Content = string.sub (HTTP_Content, chunkStart + chunkLen + 1)
			end
		end

		HTTP_Content = result
	end

	if (rcvHeaders ['CONTENT-ENCODING'] and rcvHeaders ['CONTENT-ENCODING'] == 'gzip') then
		HTTP_Content = Helper.Deflate (HTTP_Content)
	end

	Buffer [idBinding] = ''

	if (HTTP_Content == nil) then HTTP_Content = '' end

	return HTTP_Content, rcvHeaders
end

function Helper.StartXMLQ ()
	if (NETWORK_BOUND == false) then return end

	dbg('Starting XML Q')
	-- Insert commands here to start Q - remeber to add XML_CONNECTION = true in tParams

	CommandsSent = {}
	NetQ = Helper.MakePermanentNetQ (6001, IP, PORT, {FAILCOUNT = 0, XML_CONNECTION = true, TERMINATOR = '\r\n'})

	NetQ.ConnectFail = function ()
		dbg ('Connection failed check IP and device power, retrying in 60 seconds')
		CommandsSent = {} -- clear commands sent Q on rebuild
		NetQ = NetQ.KillQ ()
		Timer.ResetIP = Helper.AddTimer (Timer.ResetIP, 60, 'SECONDS', false)
	end

	NetQ.ConnectStall = function ()
		NetQ.ConnectFail ()
	end

end
function Helper.UpdateDriver (filename, index)

	if (not (filename and index)) then
		return
	end

	if (not Helper.VersionCheck21 ()) then
		Helper.DriverInfo ('Auto update requires C4 2.1 or greater')
		return
	end

	local drivers = C4:GetDevicesByC4iName (filename)

	for id, _ in pairs (drivers) do
		C4:SendToDevice (id, 'CANCEL_UPDATE_TIMER', {})
	end

	if (Helper.VersionCheck20 ()) then
		UpdateQ = Helper.MakeTransientNetQ (6510 + index, 'www.extravegetables.com', 443, {SSL = true, FAILCOUNT = 0, NAME = 'EV Update Server'})
	else
		UpdateQ = Helper.MakeTransientNetQ (6510 + index, 'www.extravegetables.com', 80, {FAILCOUNT = 0, NAME = 'EV Update Server'})
	end

	local remoteversion

	local updatesremaining = 2

	UpdateCommands = {}

	UpdateQ.ProcessMessage = function ()
		local content, headers = Helper.ProcessHTTP (6510 + index)
		local command = table.remove (UpdateCommands, 1)

		if (content and headers and headers ['HTTP/1.1'] and string.sub (headers ['HTTP/1.1'], 1, 3) == '200') then
			if (command == 'VER') then

				C4:FileSetDir ('/etc/c4i')
				local fileOnDisk = Helper.FileRead (filename) or ''
				C4:FileSetDir ('')
				remoteversion = content
				local diskver = string.match (fileOnDisk, '<category name="version" description="Driver Version">(.-)</category>')

				local loc, rem, dsk = {}, {}, {}

				loc [1], loc [2], loc [3] = string.match ((VERSION or ''), '(%d+)\.(%d+)\.(%d+)')
				rem [1], rem [2], rem [3] = string.match ((remoteversion or ''), '(%d+)\.(%d+)\.(%d+)')
				dsk [1], dsk [2], dsk [3] = string.match ((diskver or ''), '(%d+)\.(%d+)\.(%d+)')

				local getDisk, getMemory
				for i = 1, 3 do
					local remnum = tonumber (rem [i]) or 0
					local dsknum = tonumber (dsk [i]) or 0
					if (remnum > dsknum) then
						getDisk = true
						break
					end
					if (remnum < dsknum) then
						getDisk = false
						break
					end
				end
				if (getDisk) then
					Helper.DriverInfo ('Getting version to store on disk')
					local msg = 'GET /updates/' .. filename .. ' HTTP/1.1\r\nHost: www.extravegetables.com\r\nConnection: close\r\n\r\n'
					table.insert (UpdateCommands, 'C4I')
					UpdateQ.AddMessage (msg)
				else
					updatesremaining = updatesremaining - 1
					Helper.DriverInfo ('Version on disk is latest available')
				end

				if (filename ~= FILENAME [1]) then
					getMemory = getDisk
				else
					for i = 1, 3 do
						local remnum = tonumber (rem [i]) or 0
						local locnum = tonumber (loc [i]) or 0
						if (remnum > locnum) then
							getMemory = true
							break
						end
						if (remnum < locnum) then
							getMemory = false
							break
						end
					end
				end

				if (getMemory) then
					if (filename == FILENAME [1]) then Helper.DriverInfo ('Getting version to load into memory') end
					local msg = 'GET /updates/' .. filename .. '.lua HTTP/1.1\r\nHost: www.extravegetables.com\r\nConnection: close\r\n\r\n'
					table.insert (UpdateCommands, 'LUA')
					UpdateQ.AddMessage (msg)
				else
					updatesremaining = updatesremaining - 1
					Helper.DriverInfo ('Version in memory is latest available')
				end

			elseif (command == 'C4I') then
				C4:FileSetDir ('/etc/c4i')
				Helper.FileWrite (filename, content, true)
				C4:FileSetDir ('')
				Helper.DriverInfo ('New version stored on disk')
				updatesremaining = updatesremaining - 1

			elseif (command == 'LUA') then
				local basedir = '/control4/drivers/lua/sandbox/'
				for id, _ in pairs (drivers) do
					C4:FileSetDir (basedir .. tostring (id))
					Helper.FileWrite (filename, content, true)
					C4:FileSetDir ('')
				end

				Helper.DriverInfo ('Storing version to load into memory')
				for id, _ in pairs (drivers) do
					C4:SendToDevice (id, 'UPDATE_DRIVER_LUA', {VERSION = remoteversion})
				end
				updatesremaining = updatesremaining - 1
			end
		else
			if (command == 'VER') then
				Helper.DriverInfo ('Could not retrieve latest version information from server')
			elseif (command == 'C4I') then
				Helper.DriverInfo ('Could not download driver to store on disk')
			elseif (command == 'LUA') then
				Helper.DriverInfo ('Could not download driver to replace in memory')
			end
		end
		if (updatesremaining == 0) then
			Helper.DriverInfo ('Update completed: ' .. filename)
			local nextindex = index - 1
			if (nextindex > 0) then
				Helper.UpdateDriver (FILENAME [nextindex], nextindex)
			end
		end
	end

	local msg = 'GET /updates/' .. filename .. '.ver HTTP/1.1\r\nHost: www.extravegetables.com\r\nConnection: close\r\n\r\n'
	table.insert (UpdateCommands, 'VER')

	UpdateQ.AddMessage (msg)
end

function Helper.UpdateDriverLua (remoteversion)
	local loc, rem = {}, {}
	loc [1], loc [2], loc [3] = string.match ((VERSION or ''), '(%d+)\.(%d+)\.(%d+)')
	rem [1], rem [2], rem [3] = string.match ((remoteversion or ''), '(%d+)\.(%d+)\.(%d+)')
	local update
	for i = 1, 3 do
		local remnum = tonumber (rem [i]) or 0
		local locnum = tonumber (loc [i]) or 0
		if (remnum > locnum) then
			update = true
			break
		end
	end

	C4:FileSetDir ('')
	local content
	if (C4:FileExists (FILENAME [1])) then
		content = Helper.FileRead (FILENAME [1])
		C4:FileDelete (FILENAME [1])
	else
		return
	end
	if (not update) then return end

	local key = string.format ('%08X%08X%08X%08X', C4:CRC32 (FILENAME [1]), C4:CRC32 (string.upper (FILENAME [1])), C4:CRC32 (string.lower (FILENAME [1])), C4:CRC32 (string.reverse (FILENAME [1])))
	local luacode = C4:TEADecrypt (C4:Base64Decode (content), key)
	luacode = string.sub (luacode, 1, -8)
	local runcode, err = loadstring (luacode)
	if (type (runcode) == 'function') then
		OnDriverDestroyed ()
		runcode ()
		Helper.DriverInfo ('New version loaded into memory')
	else
		print (type (runcode), err)
	end
end

function Helper.URLDecode (s)
	s = string.gsub (s, '+', ' ')
	s = string.gsub (s, '%%(%x%x)', function (h)
										return string.char (tonumber (h, 16))
									end)
	return s
end

function Helper.URLEncode (s)
	s = string.gsub (s, '([&=+%c])', function (c)
										return string.format ('%%%02X', string.byte (c))
									end)
	s = string.gsub (s, ' ', '+')
	return s
end

function Helper.ValidateIP (address)
	local octet = {}
	octet [1], octet [2], octet [3], octet [4] = string.match (address, '^(%d?%d?%d)%.(%d?%d?%d)%.(%d?%d?%d)%.(%d?%d?%d)$')

	for i = 1, 4 do
		octet [i] = tonumber (octet [i])
		if (not (octet [i] and octet [i] >= 0 and octet [i] <= 255)) then
			Helper.DriverInfo ('Invalid IP address')
			return false
		end
	end
	Helper.DriverInfo ('Valid IP address')
	return true
end

function Helper.VersionCheck20 ()
	return (Helper.VersionCheck ('2.0.0.80198'))
end

function Helper.VersionCheck21 ()
	return (Helper.VersionCheck ('2.1.0.111282'))
end

function Helper.VersionCheck22 ()
	return (Helper.VersionCheck ('2.2.0.158343'))
end

function Helper.VersionCheck23 ()
	return (Helper.VersionCheck ('2.3.0.210826'))
end

function Helper.VersionCheck (requires_version)
	local curver = {}
	curver [1], curver [2], curver [3], curver [4] = string.match (C4:GetVersionInfo ().version, '(%d+)\.(%d+)\.(%d+)\.(%d+)')
	local reqver = {}
	reqver [1], reqver [2], reqver [3], reqver [4] = string.match (requires_version, '(%d+)\.(%d+)\.(%d+)\.(%d+)')

	for i = 1, 4 do
		local cur = tonumber (curver [i])
		local req = tonumber (reqver [i])
		if (cur > req) then
			return true
		end
		if (cur < req) then
			return false
		end
	end
	return true
end

function Helper.XMLDecode (s)

	s = string.gsub (s, '\&amp\;'	, '&')
	s = string.gsub (s, '\&amp\;'	, '&')
	s = string.gsub (s, '%<%!%[CDATA%[(.-)%]%]%>', function (a) return (a) end)
	s = string.gsub (s, '\&quot\;'	, '"')
	s = string.gsub (s, '\&lt\;'	, '<')
	s = string.gsub (s, '\&gt\;'	, '>')
	s = string.gsub (s, '\&apos\;'	, '\'')
	s = string.gsub (s, '&#x(.-);', function (a) return string.char (tonumber (a, 16) % 256) end )
	s = string.gsub (s, '&#(.-);', function (a) return string.char (tonumber (a) % 256) end )

	return s
end

function Helper.XMLTag (strName, tParams)
	local strXML = ''
	if (strName) then
		strXML = '<' .. tostring (strName) .. '>'
	end
	if (type (tParams) == 'table') then
		for k, v in pairs (tParams) do
			if (v == nil) then v = '' end
			strXML = strXML .. '<' .. tostring (k) .. '>' .. tostring (v) .. '</' .. tostring (k) .. '>'
		end
	elseif (tParams ~= nil) then
		strXML = strXML .. tostring (tParams)
	end
	if (strName) then
		strXML = strXML .. '</' .. tostring (strName) .. '>'
	end
	return strXML
end

function Hikvision.CreateMessage (code)
	cmd = "wps-cgi/RemoteControl.cgi?Cmd=" .. code .. "&Status=PressandRelease"
	local url = URL.BuildHTTP(cmd)
	--print (url)
	local ticket = URL.GetRequest(url)
	gTickets[ticket] = 'RC'

end
function Hikvision.InitialConnect ()
	cmd = "wps-cgi/RemoteControl.cgi?action=getNetworkConfig"
	local url = URL.BuildHTTP(cmd)
	local ticket = URL.GetRequest(url)
	gTickets[ticket] = 'Network'




end
function Hikvision.SetChannel (channel)
	for digit in string.gmatch(channel, "%d") do
		Hikvision.CreateMessage (CMDS ['NUMBER_'..digit])
	end
end

function Dec2Hex(nValue, length)
	if (not length) then length = 1 end
	if type(nValue) == "string" then
		nValue = tonumber(nValue);
	end
	nHexVal = string.format("%X", nValue);  -- %X returns uppercase hex, %x gives lowercase letters
	local chars = length - string.len(nHexVal)
	if (chars < 0) then chars = 0 end
	nHexVal = string.rep("0", chars) .. nHexVal
	sHexVal = nHexVal.."";
	return sHexVal;
end

function URL.AuthHeader()

	local http_headers = {}

	if gAuthRequired then
		if gAauthType == "BASIC" then
			local http_header_field = gUsername .. ":" .. gPassword
			http_header_field = "Basic " .. C4:Base64Encode(http_header_field)
			Dbg:Trace("HTTP Authentication: " .. http_header_field)

			http_headers["Authorization"] = http_header_field
		end
	end

	return http_headers
end

--[[
	Perform URL 'Get" request using authentication if required.
	Returns ticket (number) that identifies response with 'ReceivedAsync' callback function.
--]]
function URL.GetRequest(url)
	local l_url = url

	if gAuthRequired and gAauthType == "DIGEST" then
		l_url = string.sub(url, 1, 7) .. gUsername .. ":" .. gPassword .. "@" .. string.sub(url, 8)
	end

	return C4:urlGet(l_url, URL.AuthHeader())
end

--[[
	Perform URL 'Post" request using authentication if required.
	Returns ticket (number) that identifies response with 'ReceivedAsync' callback function.
--]]
function URL.PostRequest(url, data)
	local l_url = url

	if gAuthRequired and gAauthType == "DIGEST" then
		l_url = string.sub(url, 1, 7) .. gUsername .. ":" .. gPassword .. "@" .. string.sub(url, 8)
	end

	return C4:urlPost(l_url, data, URL.AuthHeader())
end

--[[
	Build HTTP request URL using global address and HTTP port variables and provided query string.
--]]
function URL.BuildHTTP(query_string)
	return "http://" .. IP .. ":" .. PORT .. "/" .. query_string
end
function RESPONSES.Info (message)
	dbg('Processing Reply')
	dbg (message)
	if (SettingUp == true) then
		SettingUp = false
		Timer.SetUp = Helper.KillTimer(Timer.SetUp)
		NVR_ID = nil
		CHANNELS = nil
		local _, _, name = string.find(message, 'Device name=(%w+)')
		_, _, NVR_ID = string.find(message, 'id=(%d+)')
		local _, _, mac = string.find(message, 'MAC address=(%w+:%w+:%w+:%w+:%w+:%w+)')
		 _, _, CHANNELS = string.find(message, 'Max channel=(%d+)')
		local _, _, software = string.find(message, 'Software Version=([a-z0-9%.%(%)%#]+)')
		if ((name == nil) or (CHANNELS == nil) or (mac == nil) or (software == nil)) then
			Helper.DriverInfo ('Invalid response from NVR')
		else
			Helper.DriverInfo ('NVR Connection established')
			C4:UpdateProperty ('NVR Name', name)
			C4:UpdateProperty ('NVR MAC', mac)
			C4:UpdateProperty ('Channels', CHANNELS)
			C4:UpdateProperty ('Software', software)
		end
	end
end
function RESPONSES.RC (data)
	if (data ~= 'CGI execution OK') then
		print (data)
	end
end

function RESPONSES.Network (data)
	local _, _, mac = string.find (data, '<MACAddress>(%w+:%w+:%w+:%w+:%w+:%w+)</MACAddress>')
	if (mac ~= nil) then
		Helper.DriverInfo ('NVR Connection established')
		C4:UpdateProperty ('NVR MAC', mac)

		cmd = "wps-cgi/RemoteControl.cgi?action=getMachineName"
		local url = URL.BuildHTTP(cmd)
		local ticket = URL.GetRequest(url)
		gTickets[ticket] = 'Machine'
	end
end

function RESPONSES.Machine (data)
	local _, _, device = string.find (data, '<DeviceName>(.+)</DeviceName>')
	print ('Name ' .. device)
	C4:UpdateProperty ('NVR Name', device)

	cmd = "wps-cgi/RemoteControl.cgi?action=getSerialNo"
	local url = URL.BuildHTTP(cmd)
	local ticket = URL.GetRequest(url)
	gTickets[ticket] = 'Serial'
end

function RESPONSES.Serial (data)
	local _, _, serial = string.find (data, '<serialNum>(.+)</serialNum>')
	print ('Serial ' .. serial)
	C4:UpdateProperty ('Serial', serial)

end